@@ -119,6 +119,7 @@ struct osmo_stream_cli {
 	int (*read_cb)(struct osmo_stream_cli *cli);
 	int (*iofd_read_cb)(struct osmo_stream_cli *cli, struct msgb *msg);
 	int (*write_cb)(struct osmo_stream_cli *cli);
+	int (*segmentation_cb)(struct msgb *msg);
 	void				*data;
 	int				flags;
 	int				reconnect_timeout;
@@ -160,7 +161,8 @@ static void osmo_stream_cli_close_iofd(struct osmo_stream_cli *cli)
 	if (!cli->iofd)
 		return;
 
-	osmo_iofd_close(cli->iofd);
+	osmo_iofd_free(cli->iofd);
+	cli->iofd = NULL;
 }
 
 static void osmo_stream_cli_close_ofd(struct osmo_stream_cli *cli)
@@ -209,7 +211,11 @@ void osmo_stream_cli_close(struct osmo_stream_cli *cli)
 	}
 }
 
-static inline int osmo_stream_cli_fd(const struct osmo_stream_cli *cli)
+/*! \brief Get file descriptor of the stream client socket
+ *  \param[in] cli Stream Client of which we want to obtain the file descriptor
+ *  \returns File descriptor or negative in case of error */
+int
+osmo_stream_cli_get_fd(const struct osmo_stream_cli *cli)
 {
 	switch (cli->mode) {
 	case OSMO_STREAM_MODE_OSMO_FD:
@@ -298,6 +304,7 @@ static int osmo_stream_cli_write(struct osmo_stream_cli *cli)
 		}
 		msgb_free(msg);
 		osmo_stream_cli_reconnect(cli);
+		return 0;
 	}
 
 	msgb_free(msg);
@@ -313,7 +320,7 @@ static int _setsockopt_nosigpipe(struct osmo_stream_cli *cli)
 #ifdef SO_NOSIGPIPE
 	int ret;
 	int val = 1;
-	ret = setsockopt(osmo_stream_cli_fd(cli), SOL_SOCKET, SO_NOSIGPIPE, (void *)&val, sizeof(val));
+	ret = setsockopt(osmo_stream_cli_get_fd(cli), SOL_SOCKET, SO_NOSIGPIPE, (void *)&val, sizeof(val));
 	if (ret < 0)
 		LOGSCLI(cli, LOGL_ERROR, "Failed setting SO_NOSIGPIPE: %s\n", strerror(errno));
 	return ret;
@@ -327,7 +334,7 @@ static void stream_cli_handle_connecting(struct osmo_stream_cli *cli, int res)
 	int error, ret = res;
 	socklen_t len = sizeof(error);
 
-	int fd = osmo_stream_cli_fd(cli);
+	int fd = osmo_stream_cli_get_fd(cli);
 	OSMO_ASSERT(fd >= 0);
 
 	if (ret < 0) {
@@ -346,7 +353,7 @@ static void stream_cli_handle_connecting(struct osmo_stream_cli *cli, int res)
 		osmo_fd_write_disable(&cli->ofd);
 
 	/* Update sockname based on socket info: */
-	osmo_sock_get_name_buf(cli->sockname, sizeof(cli->sockname), osmo_stream_cli_fd(cli));
+	osmo_sock_get_name_buf(cli->sockname, sizeof(cli->sockname), osmo_stream_cli_get_fd(cli));
 
 	LOGSCLI(cli, LOGL_INFO, "connection established\n");
 	cli->state = STREAM_CLI_STATE_CONNECTED;
@@ -419,6 +426,7 @@ struct osmo_stream_cli *osmo_stream_cli_create(void *ctx)
 	cli->state = STREAM_CLI_STATE_CLOSED;
 	osmo_timer_setup(&cli->timer, cli_timer_cb, cli);
 	cli->reconnect_timeout = 5;	/* default is 5 seconds. */
+	cli->segmentation_cb = NULL;
 	INIT_LLIST_HEAD(&cli->tx_queue);
 
 	cli->ma_pars.sctp.version = 0;
@@ -432,6 +440,7 @@ static void stream_cli_iofd_read_cb(struct osmo_io_fd *iofd, int res, struct msg
 
 	switch (cli->state) {
 	case STREAM_CLI_STATE_CONNECTING:
+		msgb_free(msg);
 		stream_cli_handle_connecting(cli, res);
 		break;
 	case STREAM_CLI_STATE_CONNECTED:
@@ -439,6 +448,8 @@ static void stream_cli_iofd_read_cb(struct osmo_io_fd *iofd, int res, struct msg
 			osmo_stream_cli_reconnect(cli);
 		else if (cli->iofd_read_cb)
 			cli->iofd_read_cb(cli, msg);
+		else
+			msgb_free(msg);
 		break;
 	default:
 		osmo_panic("%s() called with unexpected state %d\n", __func__, cli->state);
@@ -464,13 +475,48 @@ static void stream_cli_iofd_write_cb(struct osmo_io_fd *iofd, int res, struct ms
 	}
 }
 
-static struct osmo_io_ops osmo_stream_cli_ioops = {
+static const struct osmo_io_ops osmo_stream_cli_ioops = {
 	.read_cb = stream_cli_iofd_read_cb,
 	.write_cb = stream_cli_iofd_write_cb,
 
 	.segmentation_cb = NULL,
 };
 
+#ifdef HAVE_LIBSCTP
+static void stream_cli_iofd_recvmsg_cb(struct osmo_io_fd *iofd, int res, struct msgb *msg, const struct msghdr *msgh)
+{
+	struct osmo_stream_cli *cli  = osmo_iofd_get_data(iofd);
+
+	res = stream_iofd_sctp_recvmsg_trailer(iofd, msg, res, msgh);
+
+	switch (cli->state) {
+	case STREAM_CLI_STATE_CONNECTING:
+		msgb_free(msg);
+		stream_cli_handle_connecting(cli, res);
+		break;
+	case STREAM_CLI_STATE_CONNECTED:
+		if (res == 0)
+			osmo_stream_cli_reconnect(cli);
+		/* Forward message to read callback, also if the connection failed. */
+		if (cli->iofd_read_cb)
+			cli->iofd_read_cb(cli, msg);
+		else
+			msgb_free(msg);
+		break;
+	default:
+		osmo_panic("%s() called with unexpected state %d\n", __func__, cli->state);
+	}
+}
+
+static const struct osmo_io_ops osmo_stream_cli_ioops_sctp = {
+	.recvmsg_cb = stream_cli_iofd_recvmsg_cb,
+	.sendmsg_cb = stream_cli_iofd_write_cb,
+
+	.segmentation_cb = NULL,
+};
+#endif
+
+
 /*! \brief Set a name on the cli object (used during logging)
  *  \param[in] cli stream_cli whose name is to be set
  *  \param[in] name the name to be set on cli
@@ -482,6 +528,15 @@ void osmo_stream_cli_set_name(struct osmo_stream_cli *cli, const char *name)
 		osmo_iofd_set_name(cli->iofd, name);
 }
 
+/*! \brief Retrieve name previously set on the cli object (see osmo_stream_cli_set_name())
+ *  \param[in] cli stream_cli whose name is to be retrieved
+ *  \returns The name to be set on cli; NULL if never set
+ */
+const char *osmo_stream_cli_get_name(const struct osmo_stream_cli *cli)
+{
+	return cli->name;
+}
+
 /*! \brief Set the remote address to which we connect
  *  \param[in] cli Stream Client to modify
  *  \param[in] addr Remote IP address
@@ -585,6 +640,30 @@ osmo_stream_cli_set_proto(struct osmo_stream_cli *cli, uint16_t proto)
 	cli->flags |= OSMO_STREAM_CLI_F_RECONF;
 }
 
+/* Configure client side segmentation for the iofd */
+static void configure_cli_segmentation_cb(struct osmo_stream_cli *cli,
+					       int (*segmentation_cb)(struct msgb *msg))
+{
+	/* Copy default settings */
+	struct osmo_io_ops client_ops;
+	osmo_iofd_get_ioops(cli->iofd, &client_ops);
+	/* Set segmentation cb for this client */
+	client_ops.segmentation_cb = segmentation_cb;
+	osmo_iofd_set_ioops(cli->iofd, &client_ops);
+}
+
+/*! \brief Set the segmentation callback for the client
+ *  \param[in,out] cli Stream Client to modify
+ *  \param[in] segmentation_cb Target segmentation callback
+ */
+void osmo_stream_cli_set_segmentation_cb(struct osmo_stream_cli *cli,
+					 int (*segmentation_cb)(struct msgb *msg))
+{
+	cli->segmentation_cb = segmentation_cb;
+	if (cli->iofd) /* Otherwise, this will be done in osmo_stream_cli_open() */
+		configure_cli_segmentation_cb(cli, segmentation_cb);
+}
+
 /*! \brief Set the socket type for the stream server link
  *  \param[in] cli Stream Client to modify
  *  \param[in] type Socket Type (like SOCK_STREAM (default), SOCK_SEQPACKET, ...)
@@ -656,9 +735,10 @@ void *osmo_stream_cli_get_data(struct osmo_stream_cli *cli)
  *  \returns Socket description or NULL in case of error */
 char *osmo_stream_cli_get_sockname(const struct osmo_stream_cli *cli)
 {
-	static char buf[OSMO_SOCK_NAME_MAXLEN];
+	static char buf[OSMO_STREAM_MAX_ADDRS * OSMO_SOCK_NAME_MAXLEN];
 
-	osmo_sock_get_name_buf(buf, OSMO_SOCK_NAME_MAXLEN, osmo_stream_cli_fd(cli));
+	osmo_sock_multiaddr_get_name_buf(buf, sizeof(buf),
+					 osmo_stream_cli_get_fd(cli), cli->proto);
 
 	return buf;
 }
@@ -809,10 +889,12 @@ void osmo_stream_cli_set_nodelay(struct osmo_stream_cli *cli, bool nodelay)
  *  \return negative on error, 0 on success */
 int osmo_stream_cli_open(struct osmo_stream_cli *cli)
 {
-	int ret, fd = -1;
+	int ret, flags;
+	int fd = -1;
+	unsigned int local_addrcnt;
 
 	/* we are reconfiguring this socket, close existing first. */
-	if ((cli->flags & OSMO_STREAM_CLI_F_RECONF) && osmo_stream_cli_fd(cli) >= 0)
+	if ((cli->flags & OSMO_STREAM_CLI_F_RECONF) && osmo_stream_cli_get_fd(cli) >= 0)
 		osmo_stream_cli_close(cli);
 
 	cli->flags &= ~OSMO_STREAM_CLI_F_RECONF;
@@ -827,11 +909,18 @@ int osmo_stream_cli_open(struct osmo_stream_cli *cli)
 		switch (cli->proto) {
 #ifdef HAVE_LIBSCTP
 		case IPPROTO_SCTP:
+			local_addrcnt = cli->local_addrcnt;
+			flags = OSMO_SOCK_F_CONNECT|OSMO_SOCK_F_NONBLOCK;
+			if (cli->local_addrcnt > 0 || cli->local_port > 0) { /* explicit bind required? */
+				flags |= OSMO_SOCK_F_BIND;
+				/* If no local addr configured, use local_addr[0]=NULL by default when creating the socket. */
+				if (cli->local_addrcnt == 0)
+					local_addrcnt = 1;
+			}
 			ret = osmo_sock_init2_multiaddr2(cli->sk_domain, cli->sk_type, cli->proto,
-							(const char **)cli->local_addr, cli->local_addrcnt, cli->local_port,
+							(const char **)cli->local_addr, local_addrcnt, cli->local_port,
 							(const char **)cli->addr, cli->addrcnt, cli->port,
-							OSMO_SOCK_F_CONNECT|OSMO_SOCK_F_BIND|OSMO_SOCK_F_NONBLOCK,
-							&cli->ma_pars);
+							flags, &cli->ma_pars);
 			break;
 #endif
 		default:
@@ -865,14 +954,30 @@ int osmo_stream_cli_open(struct osmo_stream_cli *cli)
 			goto error_close_socket;
 		break;
 	case OSMO_STREAM_MODE_OSMO_IO:
-		if (!cli->iofd)
-			cli->iofd = osmo_iofd_setup(cli, fd, cli->name, OSMO_IO_FD_MODE_READ_WRITE, &osmo_stream_cli_ioops, cli);
+		/* Be sure that previous osmo_io instance is freed before creating a new one. */
+		osmo_stream_cli_close_iofd(cli);
+#ifdef HAVE_LIBSCTP
+		if (cli->proto == IPPROTO_SCTP) {
+			cli->iofd = osmo_iofd_setup(cli, fd, cli->name, OSMO_IO_FD_MODE_RECVMSG_SENDMSG,
+						    &osmo_stream_cli_ioops_sctp, cli);
+			if (cli->iofd)
+				osmo_iofd_set_cmsg_size(cli->iofd, CMSG_SPACE(sizeof(struct sctp_sndrcvinfo)));
+		} else {
+#else
+		if (true) {
+#endif
+			cli->iofd = osmo_iofd_setup(cli, fd, cli->name, OSMO_IO_FD_MODE_READ_WRITE,
+						    &osmo_stream_cli_ioops, cli);
+		}
 		if (!cli->iofd)
 			goto error_close_socket;
 
+		osmo_iofd_notify_connected(cli->iofd);
+
+		configure_cli_segmentation_cb(cli, cli->segmentation_cb);
+
 		if (osmo_iofd_register(cli->iofd, fd) < 0)
 			goto error_close_socket;
-		osmo_iofd_notify_connected(cli->iofd);
 		break;
 	default:
 		OSMO_ASSERT(false);
@@ -902,16 +1007,31 @@ static void cli_timer_cb(void *data)
  *  \param[in] msg Message buffer to enqueue in transmit queue */
 void osmo_stream_cli_send(struct osmo_stream_cli *cli, struct msgb *msg)
 {
+	int rc;
+
 	OSMO_ASSERT(cli);
 	OSMO_ASSERT(msg);
 
+	if (!osmo_stream_cli_is_connected(cli)) {
+		LOGSCLI(cli, LOGL_ERROR, "send: not connected, dropping data!\n");
+		msgb_free(msg);
+		return;
+	}
+
 	switch (cli->mode) {
 	case OSMO_STREAM_MODE_OSMO_FD:
 		msgb_enqueue(&cli->tx_queue, msg);
 		osmo_fd_write_enable(&cli->ofd);
 		break;
 	case OSMO_STREAM_MODE_OSMO_IO:
-		osmo_iofd_write_msgb(cli->iofd, msg);
+		/* whenever osmo_stream_cli_is_connected() [see above check], we should have an iofd */
+		OSMO_ASSERT(cli->iofd);
+		if (cli->proto == IPPROTO_SCTP)
+			rc = stream_iofd_sctp_send_msgb(cli->iofd, msg, MSG_NOSIGNAL);
+		else
+			rc = osmo_iofd_write_msgb(cli->iofd, msg);
+		if (rc < 0)
+			msgb_free(msg);
 		break;
 	default:
 		OSMO_ASSERT(false);
